#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""B-ASIC Scheduler-gui Graphics Graph Item Module.

Contains the scheduler-gui GraphicsGraphItem class for drawing and
maintain a component in a graph.
"""
import os
import sys
from typing     import Any, Optional
from pprint     import pprint
from typing     import Any, AnyStr, Generic, Protocol, TypeVar, Union, Optional, overload, Final, final, List
# from typing_extensions import Self, Final, Literal, LiteralString, TypeAlias, final
import numpy    as np
from copy       import deepcopy
from itertools  import combinations

import qtpy
from qtpy import QtCore
from qtpy import QtGui
from qtpy import QtWidgets

# QGraphics and QPainter imports
from qtpy.QtCore    import (
    Qt, QObject, QRect, QRectF, QPoint, QSize, QSizeF, QByteArray, Slot)
from qtpy.QtGui     import (
    QPaintEvent, QPainter, QPainterPath, QColor, QBrush, QPen, QFont, QPolygon, QIcon, QPixmap,
    QLinearGradient, QTransform)
from qtpy.QtWidgets import (
    QApplication,
    QGraphicsView, QGraphicsScene, QGraphicsWidget,
    QGraphicsLayout, QGraphicsLinearLayout, QGraphicsGridLayout, QGraphicsLayoutItem, QGraphicsAnchorLayout,
    QGraphicsItem, QGraphicsItemGroup, QGraphicsPathItem, QGraphicsLineItem, QGraphicsRectItem,
    QStyleOptionGraphicsItem, QWidget, QGraphicsObject,
    QGraphicsSceneContextMenuEvent, QGraphicsSceneHoverEvent)
from qtpy.QtCore    import (
    QPoint, QPointF, QEvent)

# B-ASIC
import logger
from b_asic.schedule            import Schedule
from b_asic.graph_component     import GraphComponent
from b_asic.special_operations  import Input, Output
from graphics_component_item    import GraphicsComponentItem
from graphics_axes_item         import GraphicsAxesItem
from graphics_graph_event       import GraphicsGraphEvent


class GraphicsGraphItem(QGraphicsItemGroup, GraphicsGraphEvent):
    """A class to represent a graph in a QGraphicsScene. This class is a
    subclass of QGraphicsItemGroup and contains the objects, axes from
    GraphicsAxesItem, as well as components from GraphicsComponentItem. It
    also inherits from GraphicsGraphEvent, which acts as a filter for events
    to GraphicsComponentItem objects."""
    _schedule:          Schedule
    _axes:              GraphicsAxesItem
    _components:        List[GraphicsComponentItem]
    _components_height: float
    _x_axis_indent:     float
    _event_items:       List[QGraphicsItem]

    
    def __init__(self, schedule: Schedule, parent: Optional[QGraphicsItem] = None):
        """Constructs a GraphicsGraphItem. 'parent' is passed to QGraphicsItemGroup's constructor."""
        super().__init__(parent)
        self._schedule = deepcopy(schedule)
        self._axes = None
        self._components = []
        self._components_height = 0.0
        self._x_axis_indent = 0.2
        self._event_items = []

        self._make_graph()
        
    def clear(self) -> None:
        """Sets all children's parent to 'None' and delete the children objects."""
        self._event_items = []
        for item in self.childItems():
            item.setParentItem(None)
            del item
    
    def is_component_valid_pos(self, pos: float) -> bool:
        """Takes in a component position and returns true if the component's new
        position is valid, false otherwise."""
        # TODO: implement
        # item = self.scene().mouseGrabberItem()
        if pos < 0:
            return False
        return True

    def is_valid_delta_time(self, delta_time: int) -> bool:
        """Takes in a delta time and returns true if the new schedule time is valid
        , false otherwise."""
        # TODO: implement
        # item = self.scene().mouseGrabberItem()
        assert self.schedule is not None , "No schedule installed."
        return self.schedule.schedule_time + delta_time >= self.schedule.get_max_end_time()
        
    
    def set_schedule_time(self, delta_time: int) -> None:
        """Set the schedule time and redraw the graph."""
        assert self.schedule is not None , "No schedule installed."
        self.schedule.set_schedule_time(self.schedule.schedule_time + delta_time)
        # scene = self.scene()
        # if scene is not None:
        #     self.removeSceneEventFilters(self._axes.event_items)
        # self._axes.update_axes(width = self._axes.width + delta_time)
        # if scene is not None:
        #     self.installSceneEventFilters(self._axes.event_items)
            # print(f'self._axes.event_items {self._axes.event_items}')
        # print(f'set_schedule_time({delta_time})')
        
        self._axes.set_width(self._axes.width + delta_time)
        # self.scene().views()[0].updateScene(QRectF(0, 0, 1, 1))
        
        
    @property
    def schedule(self) -> Schedule:
        return self._schedule
    
    @property
    def axes(self) -> GraphicsAxesItem:
        return self._axes
    
    @property
    def components(self) -> list[GraphicsComponentItem]:
        return self._components
    
    @property
    def event_items(self) -> List[QGraphicsItem]:
        """Returnes a list of objects, that receives events."""
        return self._event_items
    
    def _make_graph(self) -> None:
        """Makes a new graph out of the stored attributes."""
        # build components
        spacing = 0.2
        # print('Start times:')
        for op_id, op_start_time in self.schedule.start_times.items():
            op = self.schedule.sfg.find_by_id(op_id)
            
            if not isinstance(op, (Input, Output)):
                self._components_height += spacing
                component = GraphicsComponentItem(op_id, op.latency_offsets, op.execution_time)
                component.setPos(self._x_axis_indent + op_start_time, self._components_height)
                self._components.append(component)
                self._components_height += component.height
                self._event_items += component.event_items
        # self._components_height += spacing

        # build axes
        schedule_time = self.schedule.schedule_time
        self._axes = GraphicsAxesItem(schedule_time, self._components_height - spacing)
        self._axes.setPos(0, self._components_height + spacing*2)
        self._event_items += self._axes.event_items
        # self._axes.width = schedule_time

        # add axes and components
        self.addToGroup(self._axes)
        # self._axes.update_axes(schedule_time - 2, self._components_height, self._x_axis_indent)
        for component in self._components:
            self.addToGroup(component)
        # self.addToGroup(self._components)